#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import unicode_literals

import sys

try:
    unicode
except NameError:
    # Python 3
    def ustr(s, encoding="utf-8"):
        if s is None:
            return ""
        elif isinstance(s, str):
            return s
        try:
            return s.decode(encoding)
        except AttributeError:
            return str(s)
else:
    # Python 2
    def ustr(s, encoding="utf-8"):
        if s is None:
            return ""
        elif isinstance(s, str):
            return s.decode(encoding)
        else:
            return unicode(s)

try:
    long
except NameError:
    # Python 3
    is_integer = lambda x: isinstance(x, int)
    is_numeric = lambda x: isinstance(x, (int, float, complex))
else:
    # Python 2
    is_integer = lambda x: isinstance(x, (int, long))
    is_numeric = lambda x: isinstance(x, (int, float, long, complex))


class TextTable(object):

    @classmethod
    def cell(cls, value, size):
        if value == "#" or is_numeric(value):
            text = ustr(value).rjust(size)
        else:
            text = ustr(value).ljust(size)
        return text

    def __init__(self, header, border=False):
        self.__header = list(map(ustr, header))
        self.__rows = []
        self.__widths = list(map(len, self.__header))
        self.__repr = None
        self.border = border

    def __repr__(self):
        if self.__repr is None:
            widths = self.__widths
            if self.border:
                lines = [
                    " " + " | ".join(self.cell(value, widths[i]) for i, value in enumerate(self.__header)) + "\n",
                    "-" + "-+-".join("-" * widths[i] for i, value in enumerate(self.__header)) + "-\n",
                ]
                for row in self.__rows:
                    lines.append(" " + " | ".join(self.cell(value, widths[i]) for i, value in enumerate(row)) + "\n")
            else:
                lines = [
                    " ".join(self.cell(value, widths[i]) for i, value in enumerate(self.__header)) + "\n",
                ]
                for row in self.__rows:
                    lines.append(" ".join(self.cell(value, widths[i]) for i, value in enumerate(row)) + "\n")
            self.__repr = "".join(lines)
            if sys.version_info < (3,):
                self.__repr = self.__repr.encode("utf-8")
        return self.__repr

    def append(self, row):
        row = list(row)
        self.__rows.append(row)
        self.__widths = [max(self.__widths[i], len(ustr(value))) for i, value in enumerate(row)]
        self.__repr = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple
from datetime import datetime

from py2neo import Service, Resource, ServiceRoot
from py2neo.util import numberise


class Monitor(Service):

    __instances = {}

    def __new__(cls, uri=None):
        """ Fetch a cached instance if one is available, otherwise create,
        cache and return a new instance.

        :param uri: URI of the cached resource
        :return: a resource instance
        """
        inst = super(Monitor, cls).__new__(cls, uri)
        return cls.__instances.setdefault(uri, inst)

    def __init__(self, uri=None):
        if uri is None:
            service_root = ServiceRoot()
            manager = Resource(service_root.resource.metadata["management"])
            monitor = Monitor(manager.metadata["services"]["monitor"])
            uri = monitor.resource.uri
        Service.__init__(self)
        self.bind(uri)

    def fetch_latest_stats(self):
        """ Fetch the latest server statistics as a list of 2-tuples, each
        holding a `datetime` object and a named tuple of node, relationship and
        property counts.
        """
        counts = namedtuple("Stats", ("node_count",
                                      "relationship_count",
                                      "property_count"))
        uri = self.resource.metadata["resources"]["latest_data"]
        latest_data = Resource(uri).get().content
        timestamps = latest_data["timestamps"]
        data = latest_data["data"]
        data = zip(
            (datetime.fromtimestamp(t) for t in timestamps),
            (counts(*x) for x in zip(
                (numberise(n) for n in data["node_count"]),
                (numberise(n) for n in data["relationship_count"]),
                (numberise(n) for n in data["property_count"]),
            )),
        )
        return data



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. note::
   This module has been designed to work on Linux and may not operate
   correctly - or at all - on other platforms.

The ``server`` module provides classes for downloading and working with Neo4j server installations.
"""


import fileinput
import re
import os
from subprocess import check_output, CalledProcessError
import shlex
import sys

from py2neo import ServiceRoot
from py2neo.env import DIST_SCHEME, DIST_HOST, NEO4J_HOME
from py2neo.packages.httpstream import download as _download
from py2neo.store import GraphStore
from py2neo.util import PropertiesParser


__all__ = ["dist_name", "dist_archive_name", "download", "GraphServer", "GraphServerProcess"]

HELP = """\
Usage: {script} «edition» «version» [«path»]

Download a Neo4j server distribution.

Report bugs to nigel@py2neo.org
"""

number_in_brackets = re.compile("\[(\d+)\]")


def dist_name(edition, version):
    """ Get the full Neo4j distribution name for the given edition and version.
    """
    return "neo4j-%s-%s" % (edition, version)


def dist_archive_name(edition, version):
    """ Get the full Neo4j archive name for the given edition and version.
    """
    return "%s-unix.tar.gz" % dist_name(edition, version)


def download(edition, version, path="."):
    """ Download the Neo4j server archive for the given edition and version.
    """
    archive_name = dist_archive_name(edition, version)
    uri = "%s://%s/%s" % (DIST_SCHEME, DIST_HOST, archive_name)
    filename = os.path.join(os.path.abspath(path), archive_name)
    _download(uri, filename)
    return filename


class GraphServerProcess(object):
    """ A Neo4j server process.
    """

    #: The server installation behind this process.
    server = None

    #: The service root exposed by this server process.
    service_root = None

    #: The PID of this process.
    pid = None

    #: The JVM arguments with which this process was started.
    jvm_arguments = None

    def __init__(self, server, service_root, **kwargs):
        self.server = server
        self.service_root = service_root
        self.pid = kwargs.get("pid")
        self.jvm_arguments = kwargs.get("jvm_arguments")

    def stop(self):
        """ Stop this server process.
        """
        self.server.stop()

    @property
    def graph(self):
        """ The graph exposed by this server process.

        :rtype: :class:`py2neo.Graph`

        """
        return self.service_root.graph


class GraphServer(object):
    """ Represents a Neo4j server installation on disk.
    """

    #: The base directory of this server installation (defaults
    #: to the value of the ``NEO4J_HOME`` environment variable).
    home = None

    __conf = None

    def __init__(self, home=NEO4J_HOME):
        self.home = home
        self.reload_conf()

    def __repr__(self):
        return "<GraphServer home=%r>" % self.home

    @property
    def conf(self):
        """ Server configuration read from Java `properties` files. This is
        returned in a (slightly modified) :class:`SafeConfigParser` instance
        and can be queried as per the standard library version of this class::

            >>> from py2neo.server import GraphServer
            >>> server = GraphServer("/home/nigel/opt/neo4j")
            >>> server.conf.get("neo4j-server", "org.neo4j.server.webadmin.data.uri")
            '/db/data/'

        """
        return self.__conf

    def reload_conf(self):
        """ Reload configuration from `properties` files.
        """
        self.__conf = PropertiesParser()
        self.conf.read_properties(os.path.join(self.home, "conf", "neo4j-server.properties"))
        tuning_properties = self.conf.get("neo4j-server", "org.neo4j.server.db.tuning.properties")
        self.conf.read_properties(os.path.join(self.home, tuning_properties))

    @property
    def script(self):
        """ The file name of the control script for this server installation.
        """
        return os.path.join(self.home, "bin", "neo4j")

    @property
    def store(self):
        """ The store for this server.

        :rtype: :class:`py2neo.store.GraphStore`

        """
        return GraphStore.for_server(self)

    def start(self):
        """ Start the server.

        :rtype: :class:`py2neo.server.GraphServerProcess`

        """
        try:
            out = check_output("%s start" % self.script, shell=True)
        except CalledProcessError as error:
            if error.returncode == 2:
                raise OSError("Another process is listening on the server port")
            elif error.returncode == 512:
                raise OSError("Another server process is already running")
            else:
                raise OSError("An error occurred while trying to start "
                              "the server [%s]" % error.returncode)
        else:
            uri = None
            kwargs = {}
            for line in out.decode("utf-8").splitlines(keepends=False):
                if line.startswith("Using additional JVM arguments:"):
                    kwargs["jvm_arguments"] = shlex.split(line[32:])
                elif line.startswith("process"):
                    numbers = number_in_brackets.search(line).groups()
                    if numbers:
                        kwargs["pid"] = int(numbers[0])
                elif line.startswith("http"):
                    uri = line.partition(" ")[0]
            if not uri:
                raise RuntimeError("Unable to parse output from server startup")
            return GraphServerProcess(self, ServiceRoot(uri), **kwargs)

    def stop(self):
        """ Stop the server.
        """
        try:
            _ = check_output(("%s stop" % self.script), shell=True)
        except CalledProcessError as error:
            raise OSError("An error occurred while trying to stop "
                          "the server [%s]" % error.returncode)

    def restart(self):
        """ Restart the server.
        """
        self.stop()
        self.start()

    @property
    def pid(self):
        """ The PID of the current executing process for this server,
        """
        try:
            out = check_output(("%s status" % self.script), shell=True)
        except CalledProcessError as error:
            if error.returncode == 3:
                return None
            else:
                raise OSError("An error occurred while trying to query the "
                              "server status [%s]" % error.returncode)
        else:
            p = None
            for line in out.decode("utf-8").splitlines(keepends=False):
                if "running" in line:
                    p = int(line.rpartition(" ")[-1])
            return p

    @property
    def info(self):
        """ Dictionary of server information.
        """
        try:
            out = check_output("%s info" % self.script, shell=True)
        except CalledProcessError as error:
            if error.returncode == 3:
                return None
            else:
                raise OSError("An error occurred while trying to fetch "
                              "server info [%s]" % error.returncode)
        else:
            data = {}
            for line in out.decode("utf-8").splitlines(keepends=False):
                try:
                    colon = line.index(":")
                except ValueError:
                    pass
                else:
                    key = line[:colon]
                    value = line[colon+1:].lstrip()
                    if key.endswith("_PORT"):
                        try:
                            value = int(value)
                        except ValueError:
                            pass
                    elif key == "CLASSPATH":
                        value = value.split(":")
                    data[key] = value
            return data

    def update_server_properties(self, **properties):
        properties = {"org.neo4j.server." + key.replace("_", "."): value
                      for key, value in properties.items()}
        conf_filename = os.path.join(self.home, "conf", "neo4j-server.properties")
        for line in fileinput.input(conf_filename, inplace=1):
            for key, value in properties.items():
                if line.startswith(key + "="):
                    line = "%s=%s\n" % (key, value)
            sys.stdout.write(line)

    @property
    def service_root(self):
        """ The service root exposed by this server.

        :return: :class:`py2neo.ServiceRoot`

        """
        return ServiceRoot("http://localhost:%s" % self.info["NEO4J_SERVER_PORT"])

    @property
    def graph(self):
        """ The graph exposed by this server.

        :return: :class:`py2neo.Graph`

        """
        return self.service_root.graph


def _help(script):
    print(HELP.format(script=os.path.basename(script)))


def main():
    script, args = sys.argv[0], sys.argv[1:]
    try:
        if args:
            if len(args) == 3:
                edition, version, path = args
                download(edition, version, path)
            elif len(args) == 2:
                edition, version = args
                download(edition, version)
            else:
                _help(script)
        else:
            _help(script)
    except Exception as error:
        sys.stderr.write(ustr(error))
        sys.stderr.write("\n")
        _help(script)
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import date, time, datetime
from decimal import Decimal
from sys import version_info

from py2neo.util import ustr


# Maximum and minimum integers supported up to Java 7.
# Java 8 also supports unsigned long which can extend
# to (2 ** 64 - 1) but Neo4j is not yet on Java 8.
minint = -2 ** 63
maxint = 2 ** 63 - 1


if version_info.major == 2:
    # Python 2
    integer_types = (int, long)
    string_types = (bytearray, bytes, str, unicode)
else:
    # Python 3
    integer_types = (int,)
    string_types = (bytearray, bytes, str)


def cast_property(value):
    """ Cast the supplied property value to something supported by
    Neo4j, raising an error if this is not possible.
    """
    if isinstance(value, (bool, float)):
        pass
    elif isinstance(value, integer_types):
        if minint <= value <= maxint:
            pass
        else:
            raise ValueError("Integer value out of range: %s" % value)
    elif isinstance(value, string_types):
        value = ustr(value)
    elif isinstance(value, (frozenset, list, set, tuple)):
        # check each item and all same type
        list_value = []
        list_type = None
        for item in value:
            item = cast_property(item)
            if list_type is None:
                list_type = type(item)
                if list_type is list:
                    raise ValueError("Lists cannot contain nested collections")
            elif not isinstance(item, list_type):
                raise TypeError("List property items must be of similar types")
            list_value.append(item)
        value = list_value
    elif isinstance(value, (datetime, date, time)):
        value = value.isoformat()
    elif isinstance(value, Decimal):
        # We'll lose some precision here but Neo4j can't
        # handle decimals anyway.
        value = float(value)
    elif isinstance(value, complex):
        value = [value.real, value.imag]
    else:
        raise TypeError("Invalid property type: %s" % type(value))
    return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.cypher.error.core import ClientError, DatabaseError, TransientError


class ConcurrentRequest(ClientError):
    """ There were concurrent requests accessing the same transaction,
    which is not allowed.
    """


class EventHandlerThrewException(ClientError):
    """ A transaction event handler threw an exception. The transaction
    will be rolled back.
    """


class InvalidType(ClientError):
    """ The transaction is of the wrong type to service the request.
    For instance, a transaction that has had schema modifications
    performed in it cannot be used to subsequently perform data
    operations, and vice versa.
    """


class UnknownId(ClientError):
    """ The request referred to a transaction that does not exist.
    """


class CouldNotBegin(DatabaseError):
    """ The database was unable to start the transaction.
    """


class CouldNotCommit(DatabaseError):
    """ The database was unable to commit the transaction.
    """


class CouldNotRollback(DatabaseError):
    """ The database was unable to roll back the transaction.
    """


class ReleaseLocksFailed(DatabaseError):
    """ The transaction was unable to release one or more of its locks.
    """


class AcquireLockTimeout(TransientError):
    """ The transaction was unable to acquire a lock, for instance due
    to a timeout or the transaction thread being interrupted.
    """


class DeadlockDetected(TransientError):
    """ This transaction, and at least one more transaction, has
    acquired locks in a way that it will wait indefinitely, and the
    database has aborted it. Retrying this transaction will most likely
    be successful.
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.core import Node, Path, Rel
from py2neo.batch.core import Batch, Job, Target


__all__ = ["PullPropertiesJob", "PullNodeLabelsJob", "PullRelationshipJob", "PullBatch"]


class PullPropertiesJob(Job):

    raw_result = True

    def __init__(self, entity):
        Job.__init__(self, "GET", Target(entity, "properties"))


class PullNodeLabelsJob(Job):

    raw_result = True

    def __init__(self, node):
        Job.__init__(self, "GET", Target(node, "labels"))


class PullRelationshipJob(Job):

    raw_result = True

    def __init__(self, relationship):
        Job.__init__(self, "GET", Target(relationship))


class PullBatch(Batch):
    """ A batch of pull jobs.
    """

    def __init__(self, graph):
        Batch.__init__(self, graph)

    def append(self, entity):
        """ Append an entity to the list to be pulled.

        :param entity: An entity such as a :class:`py2neo.Node`.

        """
        if isinstance(entity, Node):
            self.jobs.append(PullPropertiesJob(entity))
            if entity.graph.supports_node_labels:
                self.jobs.append(PullNodeLabelsJob(entity))
        elif isinstance(entity, Rel):
            self.jobs.append(PullRelationshipJob(entity))
        elif isinstance(entity, Path):
            for relationship in entity.relationships:
                self.jobs.append(PullRelationshipJob(relationship))
        else:
            raise TypeError("Cannot pull object of type %s" % entity.__class__.__name__)

    def pull(self):
        """ Pull details to all entities in this batch from their
        remote counterparts.
        """
        for i, result in enumerate(self.graph.batch.submit(self)):
            job = self.jobs[i]
            entity = job.target.entity
            if isinstance(job, PullPropertiesJob):
                entity.properties.replace(result.content)
            elif isinstance(job, PullNodeLabelsJob):
                entity.labels.replace(result.content)
            elif isinstance(job, PullRelationshipJob):
                entity.__class__.hydrate(result.content, entity)
